﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace phX.Framework
{
    // ref. http://www.vcskicks.com/3d_gdiplus_drawing.php

    public static class Math3D
    {
        public class Vector2D
        {
            public double X;
            public double Y;

            public Vector2D(int x, int y)
            {
                X = x;
                Y = y;
            }

            public Vector2D(float x, float y)
            {
                X = (double)x;
                Y = (double)y;
            }

            public Vector2D(double x, double y)
            {
                X = x;
                Y = y;
            }

            public Vector2D()
            {
            }

            public override string ToString()
            {
                return "(" + X.ToString() + ", " + Y.ToString() + ")";
            }
        }

        public class Vector3D
        {
            public double X;
            public double Y;
            public double Z;

            public Vector3D(int x, int y, int z)
            {
                X = x;
                Y = y;
                Z = z;
            }

            public Vector3D(float x, float y, float z)
            {
                X = (double)x;
                Y = (double)y;
                Z = (double)z;
            }

            public Vector3D(double x, double y, double z)
            {
                X = x;
                Y = y;
                Z = z;
            }

            public Vector3D()
            {
            }

            public override string ToString()
            {
                return "(" + X.ToString() + ", " + Y.ToString() + ", " + Z.ToString() + ")";
            }
        }

        public class Camera
        {
            //For 3D drawing we need a vector of perspective, thus the Camera
            public Vector3D Position = new Vector3D();
        }

        public static Vector3D RotateX(Vector3D vector3D, double degrees)
        {
            //Here we use Euler's matrix formula for rotating a 3D vector x degrees around the x-axis

            //[ a  b  c ] [ x ]   [ x*a + y*b + z*c ]
            //[ d  e  f ] [ y ] = [ x*d + y*e + z*f ]
            //[ g  h  i ] [ z ]   [ x*g + y*h + z*i ]

            //[ 1    0        0   ]
            //[ 0   cos(x)  sin(x)]
            //[ 0   -sin(x) cos(x)]

            double cDegrees = (Math.PI * degrees) / 180.0f; //Convert degrees to radian for .Net Cos/Sin functions
            double cosDegrees = Math.Cos(cDegrees);
            double sinDegrees = Math.Sin(cDegrees);

            double y = (vector3D.Y * cosDegrees) + (vector3D.Z * sinDegrees);
            double z = (vector3D.Y * -sinDegrees) + (vector3D.Z * cosDegrees);

            return new Vector3D(vector3D.X, y, z);
        }

        public static Vector3D RotateY(Vector3D vector3D, double degrees)
        {
            //Y-axis

            //[ cos(x)   0    sin(x)]
            //[   0      1      0   ]
            //[-sin(x)   0    cos(x)]

            double cDegrees = (Math.PI * degrees) / 180.0; //Radians
            double cosDegrees = Math.Cos(cDegrees);
            double sinDegrees = Math.Sin(cDegrees);

            double x = (vector3D.X * cosDegrees) + (vector3D.Z * sinDegrees);
            double z = (vector3D.X * -sinDegrees) + (vector3D.Z * cosDegrees);

            return new Vector3D(x, vector3D.Y, z);
        }

        public static Vector3D RotateZ(Vector3D vector3D, double degrees)
        {
            //Z-axis

            //[ cos(x)  sin(x) 0]
            //[ -sin(x) cos(x) 0]
            //[    0     0     1]

            double cDegrees = (Math.PI * degrees) / 180.0; //Radians
            double cosDegrees = Math.Cos(cDegrees);
            double sinDegrees = Math.Sin(cDegrees);

            double x = (vector3D.X * cosDegrees) + (vector3D.Y * sinDegrees);
            double y = (vector3D.X * -sinDegrees) + (vector3D.Y * cosDegrees);

            return new Vector3D(x, y, vector3D.Z);
        }

        public static Vector3D Translate(Vector3D vectors3D, Vector3D translate)
        {
            //Moves a 3D vector based on a moved reference vector
            vectors3D.X += translate.X;
            vectors3D.Y += translate.Y;
            vectors3D.Z += translate.Z;
            return vectors3D;
        }

        public static Vector3D Translate(Vector3D vectors3D, Vector3D oldOrigin, Vector3D newOrigin)
        {
            //Moves a 3D vector based on a moved reference vector
            Vector3D difference = new Vector3D(newOrigin.X - oldOrigin.X, newOrigin.Y - oldOrigin.Y, newOrigin.Z - oldOrigin.Z);
            vectors3D.X += difference.X;
            vectors3D.Y += difference.Y;
            vectors3D.Z += difference.Z;
            return vectors3D;
        }

        //These are to make the above functions workable with arrays of 3D vectors
        public static Vector3D[] RotateX(Vector3D[] vectors3D, double degrees)
        {
            for (int i = 0; i < vectors3D.Length; i++)
            {
                vectors3D[i] = RotateX(vectors3D[i], degrees);
            }
            return vectors3D;
        }

        public static Vector3D[] RotateY(Vector3D[] vectors3D, double degrees)
        {
            for (int i = 0; i < vectors3D.Length; i++)
            {
                vectors3D[i] = RotateY(vectors3D[i], degrees);
            }
            return vectors3D;
        }

        public static Vector3D[] RotateZ(Vector3D[] vectors3D, double degrees)
        {
            for (int i = 0; i < vectors3D.Length; i++)
            {
                vectors3D[i] = RotateZ(vectors3D[i], degrees);
            }
            return vectors3D;
        }

        public static Vector3D[] Translate(Vector3D[] vectors3D, Vector3D translate)
        {
            for (int i = 0; i < vectors3D.Length; i++)
            {
                vectors3D[i] = Translate(vectors3D[i], translate);
            }
            return vectors3D;
        }

        public static Vector3D[] Translate(Vector3D[] vectors3D, Vector3D oldOrigin, Vector3D newOrigin)
        {
            for (int i = 0; i < vectors3D.Length; i++)
            {
                vectors3D[i] = Translate(vectors3D[i], oldOrigin, newOrigin);
            }
            return vectors3D;
        }

        public static Vector3D[] DepthSort(Vector3D[] vectors3D)
        {
            var sortedVectors3D = vectors3D.OrderBy(i => i.Z).ToArray();
            return sortedVectors3D;
        }
    }
}
