﻿using System;
using System.Drawing;
using System.IO;
using Shared;
using System.Collections.Generic;

namespace ImageMaker
{
	public static class ImageMaker
	{
		public static void CreateMode1Raw(
			string rawFolder,
			string compiledFolder,
			string bmpFilename,
            bool formatAsCpcScreen, 
            int cpcScreenWidthInBytes,
            Palette palette,
            int removeLastBytesCount, // when R13!=0
            int R9=7            
            )
		{
			var bitmap = new Bitmap(rawFolder+bmpFilename);

			//bitmap = DivideWidthBy2IfNecessary(ref bitmap);

            if (palette == null)
            {
                palette = new Palette();
            }

			var cpcRaw = new byte[(bitmap.Width/4)*bitmap.Height];

            ushort size = (ushort)((R9+1)*0x800);

            if(formatAsCpcScreen)
            {
                cpcRaw = new byte[size - removeLastBytesCount];
            }

			int offset = 0;
			for(int y=0; y<bitmap.Height; y++)
			{
                if(formatAsCpcScreen)
                {
                    offset = (y / (R9+1)) * cpcScreenWidthInBytes;
                    offset += (y % (R9 + 1)) * 0x800;
                }

				for (int x=0; x<bitmap.Width; x+=4)
				{
					Color p0 = bitmap.GetPixel(x, y);
					Color p1 = bitmap.GetPixel(x+1, y);
					Color p2 = bitmap.GetPixel(x+2, y);
					Color p3 = bitmap.GetPixel(x+3, y);

					var paletteEntry0 = palette.AddColor(p0);
					var paletteEntry1 = palette.AddColor(p1);
					var paletteEntry2 = palette.AddColor(p2);
					var paletteEntry3 = palette.AddColor(p3);

					var cpcByte = PixelHelper.ToMode1Byte(
						paletteEntry0.IndexInPalette, 
						paletteEntry1.IndexInPalette, 
						paletteEntry2.IndexInPalette, 
						paletteEntry3.IndexInPalette);

                    if (offset < (size - removeLastBytesCount))
                    {
                        cpcRaw[offset] = cpcByte;
                    }
                    offset++;
				}
			}

			DirHelper.CreateDirectory(compiledFolder);

			File.WriteAllBytes(compiledFolder+bmpFilename+".cpcRaw", cpcRaw);

			var upperName = new FileInfo(bmpFilename).Name.Replace(".bmp","").Replace(".png","").ToUpper();

			File.WriteAllText(compiledFolder+bmpFilename+".palette.table.z80", palette.ToZ80TableString());
			File.WriteAllText(compiledFolder+bmpFilename+".palette.macro.z80", palette.ToZ80MacroString(upperName));
			File.WriteAllBytes(compiledFolder+bmpFilename+".palette.firmware.bin", palette.ToFirmwareByteArray());
			File.WriteAllBytes(compiledFolder+bmpFilename+".palette.gatearray.bin", palette.ToGateArrayByteArray());
		}

        public static void ConvertRawToVRAM(
            string rawFolder,
            string compiledFolder,
            string rawFilename,
            int widthBytes,
            int r9, // default is 7
            int startPage, // 0x00, 0x40, 0x80, 0xC0
            int pageOffset // default is 0x00 .. could be 0x20 (0x40/2) to avoid INT handler..
            )
        {
            var rawBytes = File.ReadAllBytes(rawFolder + rawFilename);

            int bmpWidth = widthBytes;
            int bmpHeight = rawBytes.Length / widthBytes;

            var cpcData = new Byte[(r9 + 1) * 0x800];

            var r12r13Data = new List<byte>();

            int VCC = 0;
            int VLC = 0;
            for (int iY = 0; iY < bmpHeight; iY++)
            {
                //if(VLC==0)
                {
                    byte R12 = (byte)((((byte)startPage / 0x40) * 0x10) + 0x0C);
                    R12 += (byte)(VLC * 0x10);
                    /*if(startPage==0x40)
                    {
                        R12 = 0x1C;
                    }
                    else if (startPage == 0x80)
                    {
                        R12 = 0x2C;
                    }
                    else if (startPage == 0xC0)
                    {
                        R12 = 0x3C;
                    }*/

                    ushort cpcOffset = (ushort)(VCC * bmpWidth);
                    byte _r12 = (byte)(R12 + ((cpcOffset / 2) >> 8));
                    byte _r13 = (byte)((cpcOffset / 2) & 255);

                    r12r13Data.Add(_r12);
                    r12r13Data.Add(_r13);
                }

                int dstPtr = (VCC * bmpWidth) + (VLC * 0x800);

                int srcPtr = (iY * bmpWidth);
                for (int iX = 0; iX < bmpWidth; iX++)
                {
                    if (cpcData[dstPtr + iX] != 0)
                    {
                        System.Diagnostics.Debugger.Break();
                    }
                    cpcData[dstPtr + iX] = rawBytes[srcPtr + iX];
                }

                VLC++;
                if (VLC == (r9 + 1))
                {
                    VLC = 0;
                    VCC++;
                }
            }

            var cpcR12R13Filename = rawFilename.Replace(".cpcRaw", ".R12R13");
            File.WriteAllBytes(compiledFolder + cpcR12R13Filename, r12r13Data.ToArray());

            var cpcVRAMFilename = rawFilename.Replace(".cpcRaw", ".cpcVRAM");
            File.WriteAllBytes(compiledFolder+cpcVRAMFilename, cpcData);

            for(VLC=0; VLC<(r9+1); VLC++)
            {
                var cpcR12R130x800Filename = rawFilename.Replace(".cpcRaw", ".cpcVRAM.VLC" + VLC);
                File.WriteAllBytes(compiledFolder + cpcR12R130x800Filename, new List<byte>(cpcData).GetRange(VLC*0x800, 0x800).ToArray());
            }
        }

        public static void CreateMode0Raw(
			string rawFolder,
			string compiledFolder,
			string bmpFilename,
            int skipPensInPalette=0)
		{
			var bitmap = new Bitmap(rawFolder+bmpFilename);

			bitmap = BitmapHelper.DivideWidthBy2IfNecessary(ref bitmap);

            var palette = new Palette(skipPensInPalette);

			var cpcRaw = new byte[((bitmap.Width+1)/2)*bitmap.Height];

			int offset = 0;
			for(int y=0; y<bitmap.Height; y++)
			{
				for (int x=0; x<bitmap.Width; x+=2)
				{
					Color p0 = bitmap.GetPixel(x, y);
                    Color p1 = p0;
                    if((x+1)<bitmap.Width)
                    {
                        p1 = bitmap.GetPixel(x+1, y);
                    }

					var paletteEntry0 = palette.AddColor(p0);
					var paletteEntry1 = palette.AddColor(p1);

					var cpcByte = PixelHelper.ToMode0Byte(paletteEntry0.IndexInPalette, paletteEntry1.IndexInPalette);
					cpcRaw[offset++] = cpcByte;
				}
			}

			DirHelper.CreateDirectory(compiledFolder);

			File.WriteAllBytes(compiledFolder+"_"+bmpFilename+".cpcRaw", cpcRaw);

			var upperName = new FileInfo(bmpFilename).Name.Replace(".bmp","").Replace(".png","").ToUpper();

			File.WriteAllText(compiledFolder+"_"+bmpFilename+".palette.table.z80", palette.ToZ80TableString());
			File.WriteAllText(compiledFolder+ "_" + bmpFilename +".palette.macro.z80", palette.ToZ80MacroString(upperName));
			File.WriteAllBytes(compiledFolder+ "_" + bmpFilename +".palette.firmware.bin", palette.ToFirmwareByteArray());
			File.WriteAllBytes(compiledFolder+ "_" + bmpFilename +".palette.gatearray.bin", palette.ToGateArrayByteArray());
		}

        public static void CreateFadeMode0Raw(
            string rawFolder,
            string compiledFolder,
            string baseBmpFilename,
            string fadeBmpFilename,
            int skipPensInPalette = 0)
        {
            var baseBitmap = new Bitmap(rawFolder + baseBmpFilename);
            var fadeBitmap = new Bitmap(rawFolder + fadeBmpFilename);

            baseBitmap = BitmapHelper.DivideWidthBy2IfNecessary(ref baseBitmap);
            fadeBitmap = BitmapHelper.DivideWidthBy2IfNecessary(ref fadeBitmap);

            var basePalette = new Palette(skipPensInPalette);
            var fadePalette = new Palette(skipPensInPalette);
            var outPalette = new Palette(skipPensInPalette);

            var cpcRaw = new byte[((baseBitmap.Width + 1) / 2) * baseBitmap.Height];

            var basePaletteIndices = new List<byte>();
            var fadePaletteIndices = new List<byte>();

            for (int y = 0; y < baseBitmap.Height; y++)
            {
                for (int x = 0; x < baseBitmap.Width; x += 2)
                {
                    Color baseP0 = baseBitmap.GetPixel(x, y);
                    Color baseP1 = baseP0;
                    if ((x + 1) < baseBitmap.Width)
                    {
                        baseP1 = baseBitmap.GetPixel(x + 1, y);
                    }

                    var paletteEntry0 = basePalette.AddColor(baseP0);
                    var paletteEntry1 = basePalette.AddColor(baseP1);
                    basePaletteIndices.Add(paletteEntry0.IndexInPalette);
                    basePaletteIndices.Add(paletteEntry1.IndexInPalette);
                }
            }

            for (int y = 0; y < fadeBitmap.Height; y++)
            {
                for (int x = 0; x < fadeBitmap.Width; x += 2)
                {
                    Color fadeP0 = fadeBitmap.GetPixel(x, y);
                    Color fadeP1 = fadeP0;
                    if ((x + 1) < fadeBitmap.Width)
                    {
                        fadeP1 = fadeBitmap.GetPixel(x + 1, y);
                    }

                    var paletteEntry0 = fadePalette.AddColor(fadeP0);
                    var paletteEntry1 = fadePalette.AddColor(fadeP1);
                    fadePaletteIndices.Add(paletteEntry0.IndexInPalette);
                    fadePaletteIndices.Add(paletteEntry1.IndexInPalette);
                }
            }

            for (int y = 0; y < fadeBitmap.Height; y++)
            {
                for (int x = 0; x < fadeBitmap.Width; x++)
                {
                    int offset = (y * fadeBitmap.Width) + x;

                    int baseIndex = basePaletteIndices[offset];
                    int fadeIndex = fadePaletteIndices[offset];

                    var basePaletteEntry = basePalette.PaletteEntries[baseIndex];
                    var fadePaletteEntry = fadePalette.PaletteEntries[fadeIndex];

                    basePaletteEntry.Color = fadePaletteEntry.Color;
                    basePaletteEntry.FirmwareColor = fadePaletteEntry.FirmwareColor;
                    basePaletteEntry.GateArrayColor = fadePaletteEntry.GateArrayColor;
                }
            }
            
            var upperName = new FileInfo(fadeBmpFilename).Name.Replace(".bmp", "").Replace(".png", "").ToUpper();

            File.WriteAllText(compiledFolder + fadeBmpFilename + ".palette.table.z80", basePalette.ToZ80TableString());
            File.WriteAllText(compiledFolder + fadeBmpFilename + ".palette.macro.z80", basePalette.ToZ80MacroString(upperName));
            File.WriteAllBytes(compiledFolder + fadeBmpFilename + ".palette.firmware.bin", basePalette.ToFirmwareByteArray());
            File.WriteAllBytes(compiledFolder + fadeBmpFilename + ".palette.gatearray.bin", basePalette.ToGateArrayByteArray());
        }

        public static void CreateShiftedTopBottom(int R2_Shift, string rawFolder, string compiledFolder, string bmpFilename)
		{
			const int screenOffsetTop = 32;
			const int screenOffsetBottom = 0;
			const int lineWidth = 96;
			const int nbCharLinesTop = 18; // line 0 to 20 are visible (was 21)
			const int nbCharLinesBottom = 17; // line 0 to 13 are visible (was 14)
			const int scanlineHeight = 8;

			var top = new byte[0x4000];
			var bottom = new byte[0x4000];

			var raw = File.ReadAllBytes(rawFolder+bmpFilename);

			DirHelper.CreateDirectory(compiledFolder);

			var topOutput = compiledFolder+bmpFilename+".top.bin";
			var bottomOutput = compiledFolder+bmpFilename+".bottom.bin";

			var srcOffset = -96*7;
			var dstOffset = screenOffsetTop;
			for(int iCharY=0; iCharY<nbCharLinesTop; iCharY++)
			{
				var lineDstOffset = dstOffset;
				for(int iScanlineY=0; iScanlineY<scanlineHeight; iScanlineY++)
				{
					var scanlineDstOffset = lineDstOffset;
					for (int iScanlineWidth=0; iScanlineWidth<lineWidth; iScanlineWidth++)
					{
						if((srcOffset>=0)&&(srcOffset<raw.Length))
						{
							top[scanlineDstOffset+(R2_Shift*4)] = raw[srcOffset];
						}
						scanlineDstOffset++;
						srcOffset++;
					}

					lineDstOffset += 0x800;
				}

				dstOffset += lineWidth + (R2_Shift*4);
			}

			dstOffset = screenOffsetBottom;
			for(int iCharY=0; iCharY<nbCharLinesBottom; iCharY++)
			{
				var lineDstOffset = dstOffset;
				for(int iScanlineY=0; iScanlineY<scanlineHeight; iScanlineY++)
				{
					var scanlineDstOffset = lineDstOffset;
					for (int iScanlineWidth=0; iScanlineWidth<lineWidth; iScanlineWidth++)
					{
						if((srcOffset>=0)&&(srcOffset<raw.Length))
						{
							bottom[scanlineDstOffset+(R2_Shift*4)] = raw[srcOffset];
						}
						scanlineDstOffset++;
						srcOffset++;
					}

					lineDstOffset += 0x800;
				}

				dstOffset += lineWidth + (R2_Shift*4);
			}

			DirHelper.CreateDirectory(new FileInfo(topOutput).Directory.FullName);
			File.WriteAllBytes(topOutput, top);
			File.WriteAllBytes(bottomOutput, bottom);
		}

		public static void CreateTopBottom(string rawFolder, string compiledFolder, string bmpFilename)
		{
			const int screenOffsetTop = 0;
			const int screenOffsetBottom = 0;
			const int lineWidth = 54*2;
			const int nbCharLinesTop = 19; // line 0 to 20 are visible
			const int nbCharLinesBottom = 16; // line 0 to 13 are visible
			const int scanlineHeight = 8;

			var top = new byte[0x4000];
			var bottom = new byte[0x4000];

			var raw = File.ReadAllBytes(rawFolder+bmpFilename);

			DirHelper.CreateDirectory(compiledFolder);

			var topOutput = compiledFolder+bmpFilename+".top.bin";
			var bottomOutput = compiledFolder+bmpFilename+".bottom.bin";

			var srcOffset = -96*7;
			var dstOffset = screenOffsetTop;
			for(int iCharY=0; iCharY<nbCharLinesTop; iCharY++)
			{
				var lineDstOffset = dstOffset;
				for(int iScanlineY=0; iScanlineY<scanlineHeight; iScanlineY++)
				{
					var scanlineDstOffset = lineDstOffset;
					for (int iScanlineWidth=0; iScanlineWidth<96; iScanlineWidth++)
					{
						if((srcOffset>=0)&&(srcOffset<raw.Length))
						{
                            if(scanlineDstOffset<16384)
                            {
                                top[scanlineDstOffset] = raw[srcOffset];
                            }
						}
						scanlineDstOffset++;
						srcOffset++;
					}

                    scanlineDstOffset += 12;
                    lineDstOffset += 0x800;
				}

				dstOffset += lineWidth;
			}

            for(int i=0; i<0x4000; i++)
            {
                bottom[i] = 0x34;
            }
            
            dstOffset = screenOffsetBottom+4;
			for(int iCharY=0; iCharY<nbCharLinesBottom; iCharY++)
			{
				var lineDstOffset = dstOffset;
				for(int iScanlineY=0; iScanlineY<scanlineHeight; iScanlineY++)
				{
					var scanlineDstOffset = lineDstOffset;
					for (int iScanlineWidth=0; iScanlineWidth<96; iScanlineWidth++)
					{
						if((srcOffset>=0)&&(srcOffset<raw.Length))
						{
                            if((scanlineDstOffset>0)&&(scanlineDstOffset < 16384-12))
                            {
                                bottom[scanlineDstOffset] = raw[srcOffset];
                            }
						}
                        
                        scanlineDstOffset++;
						srcOffset++;
					}

                    scanlineDstOffset += 12;
                    lineDstOffset += 0x800;
				}

				dstOffset += lineWidth;
			}

			DirHelper.CreateDirectory(new FileInfo(topOutput).Directory.FullName);
			File.WriteAllBytes(topOutput, top);
			File.WriteAllBytes(bottomOutput, bottom);

            var topArray = new List<byte>(top);
            var bottomArray = new List<byte>(bottom);

            for (int iVLC=0; iVLC<8; iVLC++)
            {
                File.WriteAllBytes(topOutput+".VLC"+iVLC, topArray.GetRange(iVLC*0x800, 0x800).ToArray());
                File.WriteAllBytes(bottomOutput + ".VLC" + iVLC, bottomArray.GetRange(iVLC * 0x800, 0x6B8).ToArray());
            }
		}

		public static void CreateTopBottomBabyR2(string rawFolder, string compiledFolder, string bmpFilename, int lineWidth=96)
		{
			int screenOffsetTop = 32;
			int screenOffsetBottom = 0;
			const int nbCharLinesTop = 21; // line 0 to 20 are visible
			const int nbCharLinesBottom = 17; // line 0 to 13 are visible
			const int scanlineHeight = 8;

			var top = new byte[0x4000];
			var bottom = new byte[0x4000];

			var raw = File.ReadAllBytes(rawFolder+bmpFilename);

			DirHelper.CreateDirectory(compiledFolder);

			var topOutput = compiledFolder+bmpFilename+".top.bin";
			var bottomOutput = compiledFolder+bmpFilename+".bottom.bin";

			var srcOffset = -lineWidth*7;

			var screenOffset = 0;//224;

			var dstOffset = screenOffsetTop + screenOffset;
			for(int iCharY=0; iCharY<nbCharLinesTop-3; iCharY++)
			{
				var lineDstOffset = dstOffset;
				for(int iScanlineY=0; iScanlineY<scanlineHeight; iScanlineY++)
				{
					var scanlineDstOffset = lineDstOffset;
					for (int iScanlineWidth=0; iScanlineWidth<lineWidth; iScanlineWidth++)
					{
						if((srcOffset>=0)&&(srcOffset<raw.Length))
						{
							var srcRaw = raw[srcOffset];

							if ((scanlineDstOffset >= 0) && (scanlineDstOffset < top.Length))
							{
								top[scanlineDstOffset] = srcRaw;
							}
						}
						scanlineDstOffset++;
						srcOffset++;
					}

					lineDstOffset += 0x800;
				}

				dstOffset += lineWidth;
			}

			dstOffset = screenOffsetBottom;

			//srcOffset -= 1500;
			//dstOffset -= screenOffset;
			
			for(int iCharY=0; iCharY<nbCharLinesBottom; iCharY++)
			{
				var lineDstOffset = dstOffset;
				for(int iScanlineY=0; iScanlineY<scanlineHeight; iScanlineY++)
				{
					var scanlineDstOffset = lineDstOffset;
					for (int iScanlineWidth=0; iScanlineWidth<lineWidth; iScanlineWidth++)
					{
						if((srcOffset>=0)&&(srcOffset<raw.Length))
						{
							if ((scanlineDstOffset >= 0) && (scanlineDstOffset < bottom.Length))
							{
								bottom[scanlineDstOffset] = raw[srcOffset];
							}
						}
						scanlineDstOffset++;
						srcOffset++;
					}

					lineDstOffset += 0x800;
				}

				dstOffset += lineWidth;
			}

			DirHelper.CreateDirectory(new FileInfo(topOutput).Directory.FullName);
			File.WriteAllBytes(topOutput, top);
			File.WriteAllBytes(bottomOutput, bottom);
		}
	}
}
